""" Testing TxPDOs with syned transmission type """

from durand import Node, Variable
from durand.datatypes import DatatypeEnum as DT

from ..adapter import MockAdapter, TxMsg, RxMsg


def test_local_config_sync():
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    node.tpdo[0].mapping = [(0x2000, 0)]
    node.tpdo[0].transmission_type = 3

    # receive PDO message after changing into Operational state
    adapter.test(
        [   TxMsg(0x702, "00"),  # boot-up message from NMT

            RxMsg(0x000, "80 00"),  # set Pre-Operational state
            RxMsg(0x000, "01 00"),  # set Operational state

            RxMsg(0x80, ""),  # 1st SYNC
            RxMsg(0x80, ""),  # 2nd SYNC
            RxMsg(0x80, ""),  # 3rd SYNC

            TxMsg(0x182, "05 00")
        ]
    )

def test_remote_config_sync():
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    node.tpdo[0].mapping = [(0x2000, 0)]

    # receive PDO message after changing into Operational state
    adapter.test(
        [   TxMsg(0x702, "00"),  # boot-up message from NMT

            RxMsg(0x602, "2F 00 18 02 00 00 00 00"),  # set transmission type to 0
            TxMsg(0x582, "60 00 18 02 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x000, "80 00"),  # set Pre-Operational state
            RxMsg(0x000, "01 00"),  # set Operational state


            RxMsg(0x80, ""),  # sending a SYNC (without changed data)
            TxMsg(0x182, "05 00"),
            RxMsg(0x80, ""),  # sending another SYNC (without changed data)
        ]
    )

    assert node.tpdo[0].transmission_type == 0

    node.object_dictionary.write(0x2000, 0, 0xAA)
    adapter.test(
        [   RxMsg(0x80, ""),  # sending a SYNC
            TxMsg(0x182, "AA 00"),
            RxMsg(0x80, ""),  # sending another SYNC (without changed data)

            RxMsg(0x602, "2F 00 18 02 01 00 00 00"),  # set transmission type to 1
            TxMsg(0x582, "60 00 18 02 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x80, ""),  # sending a SYNC
            TxMsg(0x182, "AA 00"),

            RxMsg(0x80, ""),  # sending a SYNC
            TxMsg(0x182, "AA 00"),
        ]
    )


from typing import TYPE_CHECKING
import logging

from durand.object_dictionary import TMultiplexor, Variable, Record, Array
from durand.datatypes import DatatypeEnum as DT
from durand.services.nmt import StateEnum
from durand import get_scheduler

if TYPE_CHECKING:
    from durand.node import Node


log = logging.getLogger(__name__)


class InhibitTimer:
    def __init__(self, duration: float):
        self._timer_id = None
        self._duration = duration
        self._retrigger_callback = None

    def trigger(self, retrigger_callback):
        if self._timer_id is None:
            self._timer_id = get_scheduler().add(self._duration, self._time_up)
        else:
            self._retrigger_callback = retrigger_callback

    def cancel(self):
        if self._timer_id is not None:
            get_scheduler().cancel(self._timer_id)

        self._timer_id = None

    def _time_up(self):
        self._timer_id = None
        callback = self._retrigger_callback
        self._retrigger_callback = None

        if callback:
            callback()

    def is_active(self):
        return self._timer_id is not None


class SyncHandler:
    """SyncHandler is handling transmission types using the SYNC protocol
    mode 0: divider is 0 and _counter will be used to detect a changed datafield
    mode 1-240: every nth sync a transmit is executed
    """

    def __init__(self, tpdo: "TPDO", divider: int):
        tpdo._node.sync.callbacks.add(self._on_sync)
        self._tpdo = tpdo
        self._counter = 0
        self._divider = divider

    def cancel(self):
        self._tpdo._node.sync.callbacks.remove(self._on_sync)

    def _on_sync(self):
        if self._divider:
            self._counter += 1
            if self._counter >= self._divider:
                self._tpdo.transmit()
        else:
            if self._counter:
                self._tpdo.transmit()
            self._counter = 0

    def update(self):
        self._counter += 1


class TPDO:
    def __init__(self, node: "Node", index: int):
        self._node = node
        self._index = index

        if index < 4:
            self._cob_id = 0x4000_0180 + (index * 0x100) + node.node_id
        else:
            self._cob_id = 0xC000_0000

        self._transmission_type = 255

        self._multiplexors = ()
        self._pack_functions = None
        self._cache = None

        self._sync_handler: SyncHandler = None
        self._inhibit_timer: InhibitTimer = None

        od = self._node.object_dictionary

        param_record = Record(name="TPDO Communication Parameter")
        param_record[1] = Variable(
            DT.UNSIGNED32, "rw", self._cob_id, name="COB-ID used by TPDO"
        )
        param_record[2] = Variable(
            DT.UNSIGNED8, "rw", self._transmission_type, name="Transmission Type"
        )
        param_record[3] = Variable(DT.UNSIGNED16, "rw", 0, name="Inhibit Time")
        od[0x1800 + index] = param_record

        od.download_callbacks[(0x1800 + index, 1)].add(self._downloaded_cob_id)
        od.download_callbacks[(0x1800 + index, 2)].add(
            self._downloaded_transmission_type
        )
        od.update_callbacks[(0x1800 + index, 3)].add(self._update_inhibit_time)

        map_var = Variable(DT.UNSIGNED32, "rw", name="Application Object")
        map_array = Array(
            map_var, length=8, mutable=True, name="TPDO Mapping Parameter"
        )
        od[0x1A00 + index] = map_array

        od.write(
            0x1A00 + index, 0, 0, downloaded=False
        )  # set number of mapped objects to 0
        od.download_callbacks[(0x1A00 + index, 0)].add(self._downloaded_map_length)

        node.nmt.state_callbacks.add(self._update_nmt_state)

    def _update_nmt_state(self, state: StateEnum):
        if state == StateEnum.OPERATIONAL:
            if self._index < 4:
                self._cob_id = (
                    (self._cob_id & 0xE000_0000)
                    + 0x180
                    + (self._index * 0x100)
                    + self._node.node_id
                )

            self._activate_mapping()
        else:
            self._deactivate_mapping()

    def _downloaded_cob_id(self, value: int):
        self._cob_id = value
        # TODO: check RTR flag to be cleared

        if value & (1 << 31):
            self._deactivate_mapping()
        else:
            self._activate_mapping()

    def _downloaded_transmission_type(self, value: int):
        self.transmission_type = value

    @property
    def transmission_type(self):
        return self._transmission_type

    @transmission_type.setter
    def transmission_type(self, value: int):
        if self._sync_handler:
            self._sync_handler.cancel()
            self._sync_handler = None

        if value <= 240:
            self._sync_handler = SyncHandler(self, value)

        self._transmission_type = value

    def _update_inhibit_time(self, value: int):
        if self._inhibit_timer:
            self._inhibit_timer.cancel()
            self._inhibit_timer = None

        if value:
            self._inhibit_timer = InhibitTimer(value * 0.000_1)  # value is [100µs]

    @property
    def enable(self) -> bool:
        return not self._cob_id & (1 << 31)

    @enable.setter
    def enable(self, value: bool):
        if value:
            self._cob_id &= ~(1 << 31)
            self._activate_mapping()
        else:
            self._cob_id |= 1 << 31
            self._deactivate_mapping()

        self._node.object_dictionary.write(
            0x1800 + self._index, 1, self._cob_id, downloaded=False
        )

    @property
    def inhibit_time(self):
        return self._node.object_dictionary.read(0x1800 + self._index, 3) * 0.000_1

    @inhibit_time.setter
    def inhibit_time(self, value: float):
        return self._node.object_dictionary.write(
            0x1800 + self._index, 3, value * 10_000, downloaded=False
        )

    def _downloaded_map_length(self, length):
        multiplexors = list()

        for subindex in range(1, length + 1):
            value = self._node.object_dictionary.read(0x1A00 + self._index, subindex)
            index, subindex = value >> 16, (value >> 8) & 0xFF
            multiplexors.append((index, subindex))

        self._map(multiplexors)

    @property
    def mapping(self):
        return self._multiplexors

    @mapping.setter
    def mapping(self, multiplexors: TMultiplexor):
        self._map(multiplexors)
        self._node.object_dictionary.write(
            0x1A00 + self._index, 0, len(multiplexors), downloaded=False
        )
        for _entry, multiplexor in enumerate(multiplexors):
            index, subindex = multiplexor
            variable = self._node.object_dictionary.lookup(index, subindex)
            value = (index << 16) + (subindex << 8) + variable.size
            self._node.object_dictionary.write(
                0x1A00 + self._index, _entry + 1, value, downloaded=False
            )

    def _map(self, multiplexors: TMultiplexor):
        self._deactivate_mapping()
        self._multiplexors = tuple(multiplexors)
        self._activate_mapping()

    def _deactivate_mapping(self):
        if self._cache is None:  # check if already deactivated
            return

        if self._inhibit_timer:
            self._inhibit_timer.cancel()

        update_callbacks = self._node.object_dictionary.update_callbacks

        for multiplexor, function in zip(self._multiplexors, self._pack_functions):
            update_callbacks[multiplexor].remove(function)

        self._cache = None
        self._pack_functions = None

    def _activate_mapping(self):
        if self._cache is not None:  # check if already activated
            return

        if self._cob_id & (1 << 31) or not self._multiplexors:
            return

        if self._node.nmt.state != StateEnum.OPERATIONAL:
            return

        self._pack_functions = []
        self._cache = []

        update_callbacks = self._node.object_dictionary.update_callbacks

        for index, multiplexor in enumerate(self._multiplexors):
            variable = self._node.object_dictionary.lookup(*multiplexor)

            def pack(value, index=index, variable=variable):
                self._cache[index] = variable.pack(value)
                if self._transmission_type == 255:
                    self.transmit()
                elif self._transmission_type == 0:
                    self._sync_handler.update()

            value = self._node.object_dictionary.read(*multiplexor)
            self._cache.append(variable.pack(value))
            self._pack_functions.append(pack)
            update_callbacks[multiplexor].add(pack)

        if self._transmission_type == 255:
            self.transmit()
        elif self._transmission_type == 0:
            self._sync_handler.update()

    def transmit(self):
        if self._inhibit_timer:
            already_active = self._inhibit_timer.is_active()
            self._inhibit_timer.trigger(self.transmit)
            if already_active:
                return

        data = b"".join(self._cache)
        self._node.adapter.send(self._cob_id & 0x1FFF_FFFF, data)



""" Testing additional SDO servers. """
from durand import Node

from ..adapter import MockAdapter, TxMsg, RxMsg


def test_remote_configure():
    """In this example an additional SDO server is remotly configured (via CAN bus)"""
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # setup sdo server 9 (rx cob: 0x640, tx_cob: 0x5c0, remote node id: 1)
    adapter.test(
        [
            RxMsg(0x602, "23 09 12 01 40 06 00 00"),  # set rx cob to 0x640
            TxMsg(0x582, "60 09 12 01 00 00 00 00"),  # receive the acknowledge
            RxMsg(0x602, "23 09 12 02 c0 05 00 00"),  # set tx cob to 0x5c0
            TxMsg(0x582, "60 09 12 02 00 00 00 00"),  # receive the acknowledge
            RxMsg(0x602, "2F 09 12 03 01 00 00 00"),  # set client node id to 1
            TxMsg(0x582, "60 09 12 03 00 00 00 00"),  # receive the acknowledge
        ]
    )

    # download object 0x1000 via SDO using SDO server 9
    adapter.test(
        [
            RxMsg(0x640, "40 00 10 00 00 00 00 00"),  # request
            TxMsg(0x5C0, "43 00 10 00 00 00 00 00"),  # response
        ]
    )

    # check if attributes are updated
    assert node.sdo_servers[9].cob_rx == 0x640
    assert node.sdo_servers[9].cob_tx == 0x5C0
    assert node.sdo_servers[9].client_node_id == 1


def test_local_configure():
    """Configuring the additional SDO server local via attributes of the SDO
    server object
    """
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # setup sdo server 9 (rx cob: 0x640, tx_cob: 0x5c0, remote node id: 1)
    node.sdo_servers[9].cob_rx = 0x640
    node.sdo_servers[9].cob_tx = 0x5C0
    node.sdo_servers[9].client_node_id = 1

    # download object 0x1000 via SDO using SDO server 9
    adapter.test(
        [
            RxMsg(0x640, "40 00 10 00 00 00 00 00"),  # request
            TxMsg(0x5C0, "43 00 10 00 00 00 00 00"),  # response
        ]
    )

    # check if object dictionary is updated
    adapter.test(
        [
            RxMsg(0x602, "40 09 12 01 00 00 00 00"),  # get rx cob
            TxMsg(0x582, "43 09 12 01 40 06 00 00"),  # receive 0x640
            RxMsg(0x602, "40 09 12 02 00 00 00 00"),  # get tx cob
            TxMsg(0x582, "43 09 12 02 c0 05 00 00"),  # receive 0x5c0
            RxMsg(0x602, "40 09 12 03 00 00 00 00"),  # get client node id
            TxMsg(0x582, "4F 09 12 03 01 00 00 00"),  # receive 1
        ]
    )



from typing import List
from unittest.mock import Mock, call

from durand.adapters import AdapterABC


class Msg:
    def __init__(self, cob_id: int, *data):
        self.cob_id = cob_id

        self._data = b""
        for part in data:
            if isinstance(part, str):
                self._data += bytes.fromhex(part)
            elif isinstance(part, int):
                self._data += bytes([part])
            elif isinstance(part, bytes):
                self._data += part
            else:
                raise ValueError(
                    "Data parts have to be string containing hexcoded bytes, bytes or integers"
                )

        if len(self._data) > 8:
            raise ValueError("Maximum 8 bytes possible (%r)" % self._data)

    @property
    def data(self):
        return self._data

    def __eq__(self, msg: "Msg"):
        return self.cob_id == msg.cob_id and self._data == msg.data


class TxMsg(Msg):
    pass


class RxMsg(Msg):
    pass


class MockAdapter(AdapterABC):
    def __init__(self):
        self.subscriptions = dict()
        self.tx_mock = Mock()

    def add_subscription(self, cob_id: int, callback):
        self.subscriptions[cob_id] = callback

    def remove_subscription(self, cob_id: int):
        self.subscriptions.pop(cob_id)

    def receive(self, cob_id: int, msg: bytes):
        """Used in tests to send a CAN message to the node"""
        callback = self.subscriptions.get(cob_id, None)
        if callback is None:
            return
        callback(cob_id, msg)

    def send(self, cob_id: int, msg: bytes):
        """When the node is sending a CAN message, it will be captured in .tx_mock"""
        self.tx_mock(cob_id, msg)

    def test(self, messages: List[Msg]):
        tx_calls = list()

        for message in messages:
            if isinstance(message, RxMsg):
                self.tx_mock.assert_has_calls(tx_calls)
                self.tx_mock.reset_mock()
                tx_calls.clear()
                self.receive(message.cob_id, message.data)
            else:
                tx_calls.append(call(message.cob_id, message.data))

        self.tx_mock.assert_has_calls(tx_calls)
        assert self.tx_mock.call_count == len(tx_calls)

        self.tx_mock.reset_mock()



""" Testing RxPDOs with syncing """

from durand import Node, Variable
from durand.datatypes import DatatypeEnum as DT

from ..adapter import MockAdapter, TxMsg, RxMsg


def test_rpdo_sync():
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    # receive PDO message after changing into Operational state
    adapter.test(
        [   TxMsg(0x702, "00"),  # boot-up message from NMT

            RxMsg(0x602, "2F 00 14 02 00 00 00 00"),  # set transmission type to 0
            TxMsg(0x582, "60 00 14 02 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x602, "23 00 16 01 02 00 00 20"),  # set mapping of 0x2000:0
            TxMsg(0x582, "60 00 16 01 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x602, "2F 00 16 00 01 00 00 00"),  # set mapping length to 1
            TxMsg(0x582, "60 00 16 00 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x000, "80 00"),  # set NMT Pre-Operational state
            RxMsg(0x000, "01 00"),  # set NMT Operational state

            RxMsg(0x202, "02 00")  # receive the PDO message (will be ignored until sync)
        ]
    )

    assert node.object_dictionary.read(0x2000, 0) == 5

    adapter.test(
        [  
            RxMsg(0x80, "")  # receive sync message
        ]
    )

    assert node.object_dictionary.read(0x2000, 0) == 2



""" Testing EMCY messages """

from durand import Node

from ..adapter import MockAdapter, TxMsg


def test_simple_emcy():
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    adapter.tx_mock.reset_mock()

    node.emcy.set(0x1000, 1, data=b'\xAA')
    node.emcy.set(0, 0)

    # receive EMCY message
    adapter.test([
            TxMsg(0x82, "00 10 01 AA 00 00 00 00"),  # receive EMCY with 0x1000 error code
            TxMsg(0x82, "00 00 00 00 00 00 00 00")  # receive reset of EMCY
        ]
    )



""" Up- and download of objects via SDO server """

from durand import Node, Variable
from durand.datatypes import DatatypeEnum as DT

from ..adapter import MockAdapter, TxMsg, RxMsg


def test_simple_access():
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000:0 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    # upload the value 1 via SDO
    adapter.test(
        [
            RxMsg(
                0x602, "2b 00 20 00 01 00 00 00"
            ),  # send the upload request with value 1
            TxMsg(0x582, "60 00 20 00 00 00 00 00"),  # receive the acknowledge
        ]
    )

    # download object via SDO
    adapter.test(
        [
            RxMsg(0x602, "40 00 20 00 00 00 00 00"),  # send the download request
            TxMsg(0x582, "4b 00 20 00 01 00 00 00"),  # received the value 1
        ]
    )


def test_nmt_state():
    """The SDO server is only active when node is in pre-operational or operational state"""
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # TODO



import struct

import pytest

from durand import Node, Variable
from durand.datatypes import DatatypeEnum as DT

from .adapter import MockAdapter, TxMsg, RxMsg


@pytest.mark.parametrize("node_id", [0x01, 0x7F])
def test_sdo_object_dictionary(node_id):
    n = Node(MockAdapter(), node_id)

    assert n.object_dictionary.lookup(0x1200, 0) == Variable(
        DT.UNSIGNED8, "const", value=2, name='Highest Sub-Index Supported'
    )
    assert n.object_dictionary.lookup(0x1200, 1) == Variable(
        DT.UNSIGNED32, "ro", value=0x600 + node_id, name='COB-ID Client->Server (rx)'
    )
    assert n.object_dictionary.lookup(0x1200, 2) == Variable(
        DT.UNSIGNED32, "ro", value=0x580 + node_id, name='COB-ID Server -> Client (tx)'
    )

    with pytest.raises(KeyError):
        n.object_dictionary.lookup(0x1200, 3)


@pytest.mark.parametrize("node_id", [0x01, 0x7F])
@pytest.mark.parametrize("index", [1, 2, 127])
def test_sdo_additional_servers(node_id, index):
    n = Node(MockAdapter(), node_id)

    assert n.object_dictionary.lookup(0x1200 + index, 0) == Variable(
        DT.UNSIGNED8, "const", value=3, name='Highest Sub-Index Supported'
    )
    assert n.object_dictionary.lookup(0x1200 + index, 1) == Variable(
        DT.UNSIGNED32, "rw", value=0x8000_0000, name='COB-ID Client->Server (rx)'
    )
    assert n.object_dictionary.lookup(0x1200 + index, 2) == Variable(
        DT.UNSIGNED32, "rw", value=0x8000_0000, name='COB-ID Server -> Client (tx)'
    )
    assert n.object_dictionary.lookup(0x1200 + index, 3) == Variable(DT.UNSIGNED8, "rw", name='Node-ID of the SDO Client')


@pytest.mark.parametrize("sdo_server", [1, 127])
@pytest.mark.parametrize("node_id", [0x01, 0x7F])
def test_config_sdo_server(node_id, sdo_server):
    adapter = MockAdapter()

    n = Node(adapter, node_id)

    adapter.test(
        [
            RxMsg(
                0x640, "40 00 12 00 00 00 00 00"
            ),  # cob_id 640h should be ignored (will be used later)
            RxMsg(
                0x600 + node_id, "40", sdo_server, "12 01 00 00 00 00"
            ),  # read sdo server rx cob id
            TxMsg(
                0x580 + node_id, "43", sdo_server, "12 01 00 00 00 80"
            ),  # should be 0x8000_0000
        ]
    )

    assert n.sdo_servers[sdo_server].cob_rx is None
    assert n.sdo_servers[sdo_server].cob_tx is None

    adapter.test(
        [
            # set rx cob of sdo server to 0x640
            RxMsg(
                0x600 + node_id, "23", sdo_server, "12 01 40 06 00 00"
            ),  # set sdo server rx cob id to 0x640
            TxMsg(
                0x580 + node_id, "60", sdo_server, "12 01 00 00 00 00"
            ),  # acknowledge
            RxMsg(
                0x640, "40 00 12 00 00 00 00 00"
            ),  # cob_id 640h should still be ignored
        ]
    )

    assert n.sdo_servers[sdo_server].cob_rx == 0x640
    assert n.sdo_servers[sdo_server].cob_tx is None

    adapter.test(
        [
            # set tx cob of sdo server to 0x5C0
            RxMsg(
                0x600 + node_id, "23", sdo_server, "12 02 C0 05 00 00"
            ),  # set sdo server tx cob id to 0x5C0
            TxMsg(
                0x580 + node_id, "60", sdo_server, "12 02 00 00 00 00"
            ),  # acknowledge
            # check sdo server
            RxMsg(
                0x640, "40 00 12 00 00 00 00 00"
            ),  # cob_id 640h should now work (read number of subindex in object 0x1200)
            TxMsg(
                0x5C0, "4F 00 12 00 02 00 00 00"
            ),  # response with number of subindex (3)
        ]
    )

    assert n.sdo_servers[sdo_server].cob_rx == 0x640
    assert n.sdo_servers[sdo_server].cob_tx == 0x5C0

    # set tx cob of sdo server to 0x5D0
    n.sdo_servers[sdo_server].cob_tx = 0x5D0

    adapter.test(
        [
            RxMsg(
                0x640, "40 00 12 00 00 00 00 00"
            ),  # cob_id 640h should now work (read number of subindex in object 0x1200)
            TxMsg(
                0x5D0, "4F 00 12 00 02 00 00 00"
            ),  # response with number of subindex (3)
        ]
    )

    assert n.sdo_servers[sdo_server].cob_rx == 0x640
    assert n.sdo_servers[sdo_server].cob_tx == 0x5D0

    adapter.test(
        [
            # disable rx of sdo server
            RxMsg(
                0x600 + node_id, "23", sdo_server, "12 01 00 00 00 80"
            ),  # set sdo server rx cob id to 0x8000_0000
            TxMsg(
                0x580 + node_id, "60", sdo_server, "12 01 00 00 00 00"
            ),  # acknowledge
            RxMsg(
                0x640, "40 00 12 00 00 00 00 00"
            ),  # cob_id 640h should now be ignored
        ]
    )

    assert n.sdo_servers[sdo_server].cob_rx is None

    adapter.test(
        [
            # re-enable rx of sdo server and set to 0x641
            RxMsg(
                0x600 + node_id, "23", sdo_server, "12 01 41 06 00 00"
            ),  # set sdo server rx cob id to 0x641
            TxMsg(
                0x580 + node_id, "60", sdo_server, "12 01 00 00 00 00"
            ),  # acknowledge
            RxMsg(
                0x640, "40 00 12 00 00 00 00 00"
            ),  # cob_id 640h should now be ignored
            RxMsg(
                0x641, "40 00 12 00 00 00 00 00"
            ),  # cob_id 641h should now work (read number of subindex in object 0x1200)
            TxMsg(
                0x5D0, "4F 00 12 00 02 00 00 00"
            ),  # response with number of subindex (3)
        ]
    )

    assert n.sdo_servers[sdo_server].cob_rx == 0x641

    # change rx cob of sdo server and set to 0x642
    n.sdo_servers[sdo_server].cob_rx = 0x642

    adapter.test(
        [
            RxMsg(
                0x640, "40 00 12 00 00 00 00 00"
            ),  # cob_id 640h should now be ignored
            RxMsg(
                0x641, "40 00 12 00 00 00 00 00"
            ),  # cob_id 641h should now be ignored
            RxMsg(
                0x642, "40 00 12 00 00 00 00 00"
            ),  # cob_id 642h should work (read number of subindex in object 0x1200)
            TxMsg(
                0x5D0, "4F 00 12 00 02 00 00 00"
            ),  # response with number of subindex (3)
        ]
    )

    assert n.sdo_servers[sdo_server].cob_rx == 0x642

    # disable tx and rx cob of sdo server
    n.sdo_servers[sdo_server].cob_tx = None
    n.sdo_servers[sdo_server].cob_rx = None

    adapter.test(
        [
            RxMsg(
                0x642, "40 00 12 00 00 00 00 00"
            ),  # cob_id 642h should now be ignored
        ]
    )

    assert n.sdo_servers[sdo_server].cob_tx is None

    # sanity check to see if other sdo servers are affected
    assert n.sdo_servers[2].cob_rx is None
    assert n.sdo_servers[2].cob_tx is None



from dataclasses import dataclass, field
from collections import defaultdict
from typing import Any, Dict, Tuple, Callable, Union
import itertools
import logging

from .datatypes import DatatypeEnum, struct_dict, is_numeric, is_float
from .callback_handler import CallbackHandler, FailMode


log = logging.getLogger(__name__)


TMultiplexor = Union[int, Tuple[int, int]]


@dataclass
class Variable:
    datatype: DatatypeEnum
    access: str
    value: Any = None
    factor: float = None
    minimum: float = None
    maximum: float = None
    name: str = None

    def __post_init__(self):
        if self.datatype not in DatatypeEnum:
            raise ValueError("Unsupported datatype")
        if self.access not in ("rw", "ro", "wo", "const"):
            raise ValueError("Invalid access type")
        if not is_numeric(self.datatype) and (
            self.maximum is not None or self.minimum is not None
        ):
            raise ValueError(
                "Minimum and Maximum not available with datatype %rs" % self.datatype
            )

    @property
    def writable(self):
        return self.access in ("wo", "rw")

    @property
    def readable(self):
        return self.access in ("ro", "rw", "const")

    @property
    def size(self) -> int:
        if is_numeric(self.datatype):
            return struct_dict[self.datatype].size

        return None  # no size available

    def pack(self, value) -> bytes:
        if not is_numeric(self.datatype):
            return bytes(value)

        if self.factor is not None:
            value = value / self.factor

        dt_struct = struct_dict[self.datatype]

        if not is_float(self.datatype):
            value = int(value)

        return dt_struct.pack(value)

    def unpack(self, data: bytes):
        if not is_numeric(self.datatype):
            return bytes(data)

        dt_struct = struct_dict[self.datatype]
        value = dt_struct.unpack(data)[0]
        if self.factor is not None:
            value *= self.factor

        return value


class Record:
    def __init__(self, name: str = None):
        self.name = name
        self._variables: Dict[int, Variable] = dict()

    def __getitem__(self, subindex: int):
        if subindex == 0:
            value = max(self._variables) if self._variables else 0
            return Variable(
                DatatypeEnum.UNSIGNED8,
                "const",
                value=value,
                name="Highest Sub-Index Supported",
            )

        return self._variables[subindex]

    def __setitem__(self, subindex: int, variable: Variable):
        self._variables[subindex] = variable

    def __iter__(self):
        variables = [(0, self[0])]
        variables += list(sorted(self._variables.items(), key=lambda n: n[0]))
        return variables.__iter__()

    def __len__(self):
        return len(self._variables) + 1

    def __contains__(self, subindex: int):
        return subindex == 0 or subindex in self._variables


class Array:
    def __init__(
        self, variable: Variable, length: int, mutable: bool = False, name: str = None
    ):
        self.name = name
        self._variable = variable
        self._mutable = mutable
        self.length = length

    def __getitem__(self, subindex: int):
        if subindex == 0:
            access = "rw" if self._mutable else "const"
            return Variable(
                DatatypeEnum.UNSIGNED8,
                access,
                value=self.length,
                name="Highest Sub-Index Supported",
            )

        if subindex > self.length:
            raise KeyError(f"Subindex {subindex} not available in array")

        return self._variable

    def __iter__(self):
        variables = [(0, self[0])]
        variables += [(i + 1, self._variable) for i in range(self.length)]
        return variables.__iter__()

    def __len__(self):
        return self.length + 1

    def __contains__(self, subindex: int):
        return subindex <= self.length


TObject = Union[Variable, Record, Array]


class ObjectDictionary:
    def __init__(self):
        self._variables: Dict[TMultiplexor, Variable] = dict()
        self._objects: Dict[int, TObject] = dict()
        self._data: Dict[TMultiplexor, Any] = dict()

        self.validate_callbacks: Dict[TMultiplexor, CallbackHandler] = defaultdict(
            lambda: CallbackHandler(fail_mode=FailMode.FIRST_FAIL)
        )
        self.update_callbacks: Dict[TMultiplexor, CallbackHandler] = defaultdict(
            CallbackHandler
        )
        self.download_callbacks: Dict[TMultiplexor, CallbackHandler] = defaultdict(
            CallbackHandler
        )
        self._read_callbacks: Dict[TMultiplexor, Callable] = dict()

    def __getitem__(self, index: int):
        try:
            return self._variables[index]
        except KeyError:
            return self._objects[index]

    def __setitem__(self, index: int, object: TObject):
        if isinstance(object, Variable):
            self._variables[index] = object
        else:
            self._objects[index] = object

    def lookup(self, index: int, subindex: int = None) -> Variable:
        try:
            return self._variables[index]
        except KeyError:
            if subindex is None:
                return self._objects[index]

            return self._objects[index][subindex]

    def write(self, index: int, subindex: int, value: Any, downloaded: bool = True):
        if index in self._variables:
            multiplexor = (index, 0)
        else:
            multiplexor = (index, subindex)

        if multiplexor in self.validate_callbacks:
            self.validate_callbacks[multiplexor].call(value)  # may raises exception

        self._data[multiplexor] = value

        if multiplexor in self.update_callbacks:
            self.update_callbacks[multiplexor].call(value)

        if not downloaded:
            return

        if multiplexor in self.download_callbacks:
            self.download_callbacks[multiplexor].call(value)

    def read(self, index: int, subindex: int):
        if index in self._variables:
            multiplexor = (index, 0)
        else:
            multiplexor = (index, subindex)

        if multiplexor in self._read_callbacks:
            return self._read_callbacks[multiplexor]()

        try:
            return self._data[multiplexor]
        except KeyError:
            variable = self.lookup(index, subindex)
            value = variable.value

            if value is None:
                value = 0 if is_numeric(variable.datatype) else b""

            return value

    def set_read_callback(self, index: int, subindex: int, callback) -> None:
        self._read_callbacks[(index, subindex)] = callback

    def __iter__(self):
        objects = itertools.chain(self._objects.items(), self._variables.items())
        return iter(sorted(objects, key=lambda n: n[0]))

    def __len__(self):
        return len(self._objects) + len(self._variables)

    def __contains__(self, index: int):
        return index in self._objects or index in self._variables



""" Testing LSS service """

from durand import Node

from ..adapter import MockAdapter, TxMsg, RxMsg


def test_global_selection():
    """Test example starts with a node with an undefined node id (0xFF).
    After the "switch state global" request to set every slave into configuration mode,
    the node id is set to 1.

    When switching back to waiting state via "switch state global" the node sends
    an boot up message (because the node id was undefined before).

    SDO requests are tested on node 1 to see, if the slave is responding.
    """
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=0xFF)

    adapter.test(
        [
            RxMsg(
                0x7E5, "04 01 00 00 00 00 00 00"
            ),  # switch state global to configuration state
            RxMsg(0x7E5, "11 01 00 00 00 00 00 00"),  # set node id to 1
            TxMsg(0x7E4, "11 00 00 00 00 00 00 00"),  # receive the acknowledge
            RxMsg(
                0x601, "40 00 10 00 00 00 00 00"
            ),  # requesting via SDO on node 1 will still be unanswered
            RxMsg(
                0x7E5, "04 00 00 00 00 00 00 00"
            ),  # switch state global back to waiting state
            TxMsg(0x701, "00"),  # slave responses with Boot Up message
            RxMsg(0x601, "40 00 10 00 00 00 00 00"),  # requesting via SDO on node 1
            TxMsg(0x581, "43 00 10 00 00 00 00 00"),  # receive the acknowledge
        ]
    )



from dataclasses import dataclass, fields
from re import match
from datetime import datetime
from typing import TYPE_CHECKING

from durand.object_dictionary import Variable, Record
from durand.datatypes import DatatypeEnum


if TYPE_CHECKING:
    from durand.node import Node


def datetime_to_time(d: datetime):
    return d.strftime("%I:%M") + ("AM" if d.hour < 12 else "PM")


def datetime_to_date(d: datetime):
    return d.strftime("%m-%d-%Y")


@dataclass
class FileInfo:
    FileName: str = "python_durand_device.eds"
    FileVersion: int = 0
    FileRevision: int = 0
    EDSVersion: str = "4.0"
    Description: str = None
    CreationTime: str = None
    CreationDate: str = None
    CreatedBy: str = None
    ModificationTime: str = None
    ModificationDate: str = None
    ModifiedBy: str = None

    def validate(self):
        if not 0 <= self.FileVersion <= 255:
            raise ValueError("FileVersion is Unsigned8")
        if not 0 <= self.FileRevision <= 255:
            raise ValueError("FileRevision is Unsigned8")
        if not match("[d].[d]", self.EDSVersion):
            raise ValueError("EDSVersion type mismatch")

        if self.CreationDate:
            try:
                datetime.strptime("%m-%d-%Y", self.CreationDate)
            except ValueError:
                raise ValueError("CreationDate format invalid")

        if self.CreationTime:
            if len(self.CreationTime) != 7 or self.CreationTime[5:] not in ("AM", "PM"):
                raise ValueError("CreationTime format invalid")

            try:
                datetime.strptime("%I:%M", self.CreationTime[:6])
            except ValueError:
                raise ValueError("CreationTime format invalid")

        if self.ModificationDate:
            try:
                datetime.strptime("%m-%d-%Y", self.ModificationDate)
            except ValueError:
                raise ValueError("ModificationDate format invalid")

        if self.ModificationTime:
            if len(self.ModificationTime) != 7 or self.ModificationTime[5:] not in (
                "AM",
                "PM",
            ):
                raise ValueError("ModificationTime format invalid")

            try:
                datetime.strptime("%I:%M", self.ModificationTime[:6])
            except ValueError:
                raise ValueError("ModificationTime format invalid")

    @property
    def content(self):
        content = "[FileInfo]\n"

        for field in fields(self):
            value = getattr(self, field.name)
            if value is None:
                continue

            content += f"{field.name}={value!s}\n"

        return content + "\n"


@dataclass
class DeviceInfo:
    VendorName: str = None

    @property
    def content(self):
        return "[DeviceInfo]\n\n"


def extract_objects(d: dict, indices: list) -> dict:
    extracted_dict = {index: obj for index, obj in d.items() if index in indices}
    for index in indices:
        d.pop(index, None)

    return extracted_dict


def describe_variable(index: int, subindex: int, variable: Variable) -> str:
    name = f"{index:04X}" + ("" if subindex is None else f"sub{subindex}")
    content = f"[{name}]\n"

    if variable.name:
        content += f"ParameterName={variable.name}\n"
    else:
        content += f"ParameterName=Variable{name}\n"

    content += "ObjectType=0x7\n"
    content += f"DataType=0x{variable.datatype}\n"
    content += f"AccessType={variable.access}\n"

    if variable.value is not None:
        content += f"DefaultValue={variable.value}\n"

    if variable.minimum is not None:
        content += f"LowLimit={variable.minimum}\n"

    if variable.maximum is not None:
        content += f"HighLimit={variable.maximum}\n"

    content += "PDOMapping=1\n\n"

    return content


def describe_object(index: int, object) -> str:
    if isinstance(object, Variable):
        return describe_variable(index, None, object)

    content = f"[{index:04X}]\nSubNumber={len(object)}\n"

    if object.name:
        content += f"ParameterName={object.name}\n"
    else:
        content += f"ParameterName=Object{index:04X}\n"

    content += (
        "ObjectType=0x8\n\n" if isinstance(object, Record) else "ObjectType=0x9\n\n"
    )

    for subindex, variable in object:
        content += describe_variable(index, subindex, variable)

    return content


def describe_section(name: str, objects: dict):
    content = f"[{name}]\nSupportedObjects={len(objects)}\n"

    for obj_nr, index in enumerate(objects):
        content += f"{obj_nr + 1}=0x{index:04X}\n"

    content += "\n"

    for index, object in objects.items():
        content += describe_object(index, object)

    return content


class EDS:
    def __init__(self, node: "Node"):
        self._node = node

        self.file_info = FileInfo()
        self.device_info = DeviceInfo()
        self.comments = ""

    @property
    def content(self):
        content = self.file_info.content

        content += self.device_info.content

        if self.comments:
            lines = self.comments.strip().splitlines()
            content += f"[Comments]\nLines={len(lines)}\n"
            for index, line in enumerate(lines):
                content += f"Line{index + 1}={line:s}\n"
            content += "\n"

        objects = dict(self._node.object_dictionary)

        mandatory_objects = extract_objects(objects, (0x1000, 0x1001, 0x1018))
        content += describe_section("MandatoryObjects", mandatory_objects)

        optional_indicies = [
            index for index in objects if index < 0x2000 or index >= 0x6000
        ]
        optional_objects = extract_objects(objects, optional_indicies)
        content += describe_section("OptionalObjects", optional_objects)

        content += describe_section("ManufacturerObjects", objects)

        return content


class EDSProvider:
    def __init__(self, node: "Node"):
        self._node = node



from typing import List
from dataclasses import dataclass

from .adapters import AdapterABC
from .object_dictionary import ObjectDictionary
from .services.sdo import SDOServer
from .services.pdo import TPDO, RPDO
from .eds import EDS
from .services.nmt import NMTService, StateEnum
from .services.lss import LSS
from .services.sync import Sync
from .services.emcy import EMCY
from .services.heartbeat import HeartbeatProducer
from .object_dictionary import Variable, Record
from .datatypes import DatatypeEnum as DT


SDO_SERVERS = 127


@dataclass
class NodeCapabilities:
    sdo_servers: int
    rpdos: int
    tpdos: int


FullNodeCapabilities = NodeCapabilities(sdo_servers=128, rpdos=512, tpdos=512)


class Node:
    def __init__(
        self,
        adapter: AdapterABC,
        node_id: int,
        od: ObjectDictionary = None,
        capabilities: NodeCapabilities = FullNodeCapabilities,
    ):

        self.adapter = adapter
        self.node_id = node_id
        od = ObjectDictionary() if od is None else od
        self.object_dictionary = od
        self.eds = EDS(self)

        self.nmt = NMTService(self)
        self.sync = Sync(self)

        self.tpdo = [TPDO(self, i) for i in range(capabilities.tpdos)]
        self.rpdo = [RPDO(self, i) for i in range(capabilities.rpdos)]

        self.sdo_servers: List[SDOServer] = list()

        assert (
            1 <= SDO_SERVERS <= 128
        ), "Number of SDO servers has to be between 1 and 128"

        for index in range(capabilities.sdo_servers):
            self.sdo_servers.append(SDOServer(self, index))

        self.heartbeat_producer = HeartbeatProducer(self)
        self.lss = LSS(self)
        self.emcy = EMCY(self)

        od[0x1000] = Variable(DT.UNSIGNED32, "ro", 0, name="Device Type")

        od[0x1008] = Variable(
            DT.VISIBLE_STRING, "ro", b"", name="Manufacturer Device Name"
        )
        od[0x1009] = Variable(
            DT.VISIBLE_STRING, "ro", b"", name="Manufacturer Hardware Version"
        )
        od[0x100A] = Variable(
            DT.VISIBLE_STRING, "ro", b"", name="Manufacturer Software Version"
        )

        # Identity object
        identity_record = Record(name="Identity Object")
        identity_record[1] = Variable(DT.UNSIGNED32, "ro", 0, name="Vendor-ID")
        identity_record[2] = Variable(DT.UNSIGNED32, "ro", 0, name="Product Code")
        identity_record[3] = Variable(DT.UNSIGNED32, "ro", 0, name="Revision Number")
        identity_record[4] = Variable(DT.UNSIGNED32, "ro", 0, name="Serial Number")

        od[0x1018] = identity_record
        self.nmt.set_state(StateEnum.PRE_OPERATIONAL)

        # EDS provider
        od[0x1021] = Variable(DT.DOMAIN, "ro", name="Store EDS")
        od[0x1022] = Variable(DT.UNSIGNED16, "ro", value=0, name="Store Format")
        od.set_read_callback(0x1021, 0, lambda: self.eds.content.encode())


MinimalNodeCapabilities = NodeCapabilities(sdo_servers=1, rpdos=4, tpdos=4)


class MinimalNode(Node):
    def __init__(self, adapter: AdapterABC, node_id: int, od: ObjectDictionary = None):
        Node.__init__(self, adapter, node_id, od, MinimalNodeCapabilities)



from typing import TYPE_CHECKING
from enum import IntEnum
import logging
import struct

from .nmt import StateEnum
from durand.scheduler import get_scheduler

if TYPE_CHECKING:
    from ..node import Node

log = logging.getLogger(__name__)


class LSSState(IntEnum):
    WAITING = 0
    CONFIGURATION = 1


class LSS:
    def __init__(self, node: "Node"):
        self._node = node

        self._state = LSSState.WAITING

        self._received_selective_address = [None] * 4
        self._remote_slave_address = [None] * 6
        self._fastscan_state = 0

        self._pending_baudrate = None
        self._change_baudrate_cb = None

        node.adapter.add_subscription(cob_id=0x7E5, callback=self.handle_msg)
        node.nmt.state_callbacks.add(self.on_nmt_state_update)

    def set_baudrate_change_callback(self, cb):
        self._change_baudrate_cb = cb

    def on_nmt_state_update(self, state: StateEnum):
        if state == StateEnum.INITIALISATION:
            self._state = LSSState.WAITING
            self._received_selective_address = [None] * 4

    def _get_own_address(self):
        lss_address = [None] * 4

        for index in range(4):
            lss_address[index] = self._node.object_dictionary.read((0x1018, index + 1))

        return lss_address

    def handle_msg(self, cob_id: int, msg: bytes):
        cs = msg[0]

        if self._state == LSSState.WAITING:
            method = LSS._waiting_cs_dict.get(cs, None)
        else:
            method = LSS._configuration_cs_dict.get(cs, None)

        if method is None:
            return

        method(self, msg)

    def cmd_switch_mode_global_configuration(self, msg: bytes):
        if msg[1] != 1:  # check if requested mode is CONFIGURATION
            return

        self._state = LSSState.CONFIGURATION

    def cmd_switch_mode_global_waiting(self, msg: bytes):
        if msg[1] != 0:  # check if requested mode is WAITING
            return

        if self._node.node_id == 0xFF and self._node.nmt.pending_node_id != 0xFF:
            self._node.nmt.reset()

        self._state = LSSState.WAITING

    def cmd_switch_mode_selective(self, msg: bytes):
        index = msg[0] - 0x40

        self._received_selective_address[index] = int.from_bytes(
            msg[1:5], "little", signed=False
        )

        if None in self._received_selective_address:
            return

        if self._received_selective_address == self._get_own_address():
            self._state = LSSState.CONFIGURATION
            self._node.adapter.send(0x7E4, b"\x44" + bytes(7))

        self._received_selective_address == [None] * 4

    def cmd_inquire_identity(self, msg: bytes):
        index = msg[0] - 0x5A
        value = self._get_own_address()[index]
        self._node.adapter.send(0x7E4, msg[:1] + value.to_bytes(4, "little") + bytes(3))

    def cmd_inquire_node_id(self, msg: bytes):
        self._node.adapter.send(
            0x7E4, b"\x5e" + self._node.node_id.to_bytes(1, "little") + bytes(6)
        )

    def cmd_configure_node_id(self, msg: bytes):
        node_id = msg[1]

        if 1 <= node_id <= 127 or node_id == 0xFF:
            self._node.nmt.pending_node_id = node_id
            result = 0
        else:
            result = 1

        self._node.adapter.send(
            0x7E4, b"\x11" + result.to_bytes(1, "little") + bytes(6)
        )

    def cmd_configure_bit_timing(self, msg: bytes):
        valid_table_entries = (0, 1, 2, 3, 4, 6, 7, 8)
        selector, index = msg[1:3]

        if (
            selector != 0
            or index not in valid_table_entries
            or self._change_baudrate_cb is None
        ):
            self._node.adapter.send(0x7E4, b"\x13\x01" + bytes(6))
            return

        self._pending_baudrate = index
        self._node.adapter.send(0x7E4, b"\x13\x00" + bytes(6))

    def cmd_activate_bit_timing(self, msg: bytes):
        delay = int.from_bytes(msg[1:3], "little") / 1000  # [seconds]

        if self._pending_baudrate is not None:
            get_scheduler().add(delay, self._change_baudrate, args=(delay,))

    def _change_baudrate(self, delay: float):
        self._change_baudrate_cb(self._pending_baudrate)
        self._pending_baudrate = None
        get_scheduler().add(delay, self._node.nmt.reset)

    def cmd_store_configuration(self, msg: bytes):
        # store configuration is not supported
        self._node.adapter.send(0x7E4, b"\x17\x01" + bytes(6))

    def cmd_identify_remote_slaves(self, msg: bytes):
        index = msg[0] - 0x46
        value = int.from_bytes(msg[1:5], "little")
        self._remote_slave_address[index] = value

        if None in self._remote_slave_address:
            return

        vendor, product, revision, serial = self._get_own_address()

        if (
            vendor == self._remote_slave_address[0]
            and product == self._remote_slave_address[1]
            and self._remote_slave_address[2]
            <= revision
            <= self._remote_slave_address[3]
            and self._remote_slave_address[4] <= serial <= self._remote_slave_address[5]
        ):

            self._node.adapter.send(0x7E4, b"\x47" + bytes(7))

        self._remote_slave_address = [None] * 6

    def cmd_identify_nonconfigured_remote_slaves(self, msg: bytes):
        if self._node.node_id == 0xFF:
            self._node.adapter.send(0x7E4, b"\x50" + bytes(7))

    def cmd_fastscan(self, msg: bytes):
        if self._node.node_id != 0xFF:
            return

        id_number, bit_checked, lss_sub, lss_next = struct.unpack("IBBB", msg[1:])

        if bit_checked == 0x80:
            self._fastscan_state = 0
            self._node.adapter.send(0x7E4, b"\x4F" + bytes(7))
            return

        if lss_sub != self._fastscan_state:
            return

        mask = ~((1 << bit_checked) - 1)
        lss_address = self._get_own_address()

        if (lss_address[lss_sub] & mask) != (id_number & mask):
            return

        self._fastscan_state = lss_next

        if bit_checked == 0 and lss_sub == 3:
            self._state = LSSState.CONFIGURATION

        self._node.adapter.send(0x7E4, b"\x4F" + bytes(7))

    _waiting_cs_dict = {
        0x04: cmd_switch_mode_global_configuration,
        0x40: cmd_switch_mode_selective,
        0x41: cmd_switch_mode_selective,
        0x42: cmd_switch_mode_selective,
        0x43: cmd_switch_mode_selective,
        0x46: cmd_identify_remote_slaves,
        0x47: cmd_identify_remote_slaves,
        0x48: cmd_identify_remote_slaves,
        0x49: cmd_identify_remote_slaves,
        0x4A: cmd_identify_remote_slaves,
        0x4B: cmd_identify_remote_slaves,
        0x4C: cmd_identify_nonconfigured_remote_slaves,
        0x51: cmd_fastscan,
    }

    _configuration_cs_dict = {
        0x04: cmd_switch_mode_global_waiting,
        0x11: cmd_configure_node_id,
        0x13: cmd_configure_bit_timing,
        0x15: cmd_activate_bit_timing,
        0x17: cmd_store_configuration,
        0x46: cmd_identify_remote_slaves,
        0x47: cmd_identify_remote_slaves,
        0x48: cmd_identify_remote_slaves,
        0x49: cmd_identify_remote_slaves,
        0x4A: cmd_identify_remote_slaves,
        0x4B: cmd_identify_remote_slaves,
        0x4C: cmd_identify_nonconfigured_remote_slaves,
        0x51: cmd_fastscan,
        0x5A: cmd_inquire_identity,
        0x5B: cmd_inquire_identity,
        0x5C: cmd_inquire_identity,
        0x5D: cmd_inquire_identity,
        0x5E: cmd_inquire_node_id,
    }



from typing import TYPE_CHECKING

from durand.object_dictionary import TMultiplexor, Variable, Record, Array
from durand.datatypes import DatatypeEnum as DT
from durand.services.nmt import StateEnum


if TYPE_CHECKING:
    from durand.node import Node


class RPDO:
    def __init__(self, node: "Node", index: int):
        self._node = node
        self._index = index

        if index < 4:
            self._cob_id = 0x0000_0200 + (index * 0x100) + node.node_id
        else:
            self._cob_id = 0x8000_0000

        self._transmission_type = 255

        self._multiplexors = ()
        self._unpack_function = None

        self._synced_msg = None

        od = self._node.object_dictionary

        param_record = Record(name="RPDO Communication Parameter")
        param_record[1] = Variable(
            DT.UNSIGNED32, "rw", self._cob_id, name="COB-ID used by RPDO"
        )
        param_record[2] = Variable(
            DT.UNSIGNED8, "rw", self._transmission_type, name="Transmission Type"
        )
        od[0x1400 + index] = param_record

        od.download_callbacks[(0x1400 + index, 1)].add(self._downloaded_cob_id)
        od.download_callbacks[(0x1400 + index, 2)].add(
            self._downloaded_transmission_type
        )

        map_var = Variable(DT.UNSIGNED32, "rw", name="Application Object")
        map_array = Array(
            map_var, length=8, mutable=True, name="RPDO Mapping Parameter"
        )
        od[0x1600 + index] = map_array

        od.write(
            0x1600 + index, 0, 0, downloaded=False
        )  # set number of mapped objects to 0
        od.download_callbacks[(0x1600 + index, 0)].add(self._downloaded_map_length)

        node.nmt.state_callbacks.add(self._update_nmt_state)

    def _update_nmt_state(self, state: StateEnum):
        if state == StateEnum.OPERATIONAL:
            if self._index < 4:
                self._cob_id = (
                    (self._cob_id & 0xE000_0000)
                    + 0x200
                    + (self._index * 0x100)
                    + self._node.node_id
                )

            self._activate_mapping()
        else:
            self._deactivate_mapping()

    def _downloaded_cob_id(self, value: int):
        self._cob_id = value

        if value & (1 << 31):
            self._deactivate_mapping()
        else:
            self._activate_mapping()

    def _downloaded_transmission_type(self, value: int):
        self.transmission_type = value

    @property
    def transmission_type(self):
        return self._transmission_type

    @transmission_type.setter
    def transmission_type(self, value: int):
        self._deactivate_mapping()
        self._transmission_type = value
        self._activate_mapping()

    @property
    def enable(self) -> bool:
        return not self._cob_id & (1 << 31)

    @enable.setter
    def enable(self, value: bool):
        if value:
            self._cob_id &= ~(1 << 31)
            self._activate_mapping()
        else:
            self._cob_id |= 1 << 31
            self._deactivate_mapping()

        self._node.object_dictionary.write(
            0x1400 + self._index, 1, self._cob_id, downloaded=False
        )

    def _downloaded_map_length(self, length):
        multiplexors = list()

        for subindex in range(1, length + 1):
            value = self._node.object_dictionary.read(0x1600 + self._index, subindex)
            index, subindex = value >> 16, (value >> 8) & 0xFF
            multiplexors.append((index, subindex))

        self._map(multiplexors)

    @property
    def mapping(self):
        return self._multiplexors

    @mapping.setter
    def mapping(self, multiplexors: TMultiplexor):
        self._map(multiplexors)
        self._node.object_dictionary.write(
            0x1600 + self._index, 0, len(multiplexors), downloaded=False
        )
        for _entry, multiplexor in enumerate(multiplexors):
            index, subindex = multiplexor
            variable = self._node.object_dictionary.lookup(index, subindex)
            value = (index << 16) + (subindex << 8) + variable.size
            self._node.object_dictionary.write(
                0x1600 + self._index, _entry + 1, value, downloaded=False
            )

    def _map(self, multiplexors: TMultiplexor):
        self._deactivate_mapping()
        self._multiplexors = tuple(multiplexors)
        self._activate_mapping()

    def _deactivate_mapping(self):
        if self._unpack_function is None:
            return

        self._unpack_function = None

        if self._on_sync in self._node.sync.callbacks:
            self._node.sync.callbacks.remove(self._on_sync)

        self._node.adapter.remove_subscription(cob_id=self._cob_id & 0x1FFF_FFFF)

    def _activate_mapping(self):
        if self._unpack_function is not None:
            return

        if self._cob_id & (1 << 31) or not self._multiplexors:
            return

        if self._node.nmt.state != StateEnum.OPERATIONAL:
            return

        variables = []
        for multiplexor in self._multiplexors:
            variable = self._node.object_dictionary.lookup(*multiplexor)
            variables.append(variable)

        def unpack(data: bytes):
            values = []
            for variable in variables:
                values.append(variable.unpack(data[: variable.size]))
                data = data[variable.size :]

            return values

        self._unpack_function = unpack

        if self._transmission_type <= 240:
            self._node.sync.callbacks.add(self._on_sync)

        self._node.adapter.add_subscription(
            cob_id=self._cob_id & 0x1FFF_FFFF, callback=self._handle_msg
        )

    def _on_sync(self):
        if self._synced_msg is None:
            return

        self._write_data(self._synced_msg)
        self._synced_msg = None

    def _handle_msg(self, cob_id: int, msg: bytes):
        if self._transmission_type <= 240:
            self._synced_msg = msg
            return

        self._write_data(msg)

    def _write_data(self, msg: bytes):
        values = self._unpack_function(msg)

        for multiplexor, value in zip(self._multiplexors, values):
            try:
                self._node.object_dictionary.write(*multiplexor, value)
            except:  # there is no possibility to response in such a case
                pass



import asyncio
from abc import ABCMeta, abstractmethod, abstractproperty
from dataclasses import dataclass
from typing import Callable, Tuple, TypeVar, Dict, Any
import functools
import threading
from sched import scheduler


TEntry = TypeVar("TEntry")  # type of scheduler entry


class AbstractScheduler(metaclass=ABCMeta):
    @abstractmethod
    def add(self, delay: float, callback, args=(), kwargs=None) -> TEntry:
        """Add a new scheduler entry.

        :param delay: time in seconds when the callback should be called
        :param callback: a function object called when time has passed
        :param args: tuple with positional arguments for the callback
        :param kwargs: dictionary with keyword arguments
        :returns: an id which can be used to cancel the scheduled callback
        """

    @abstractmethod
    def cancel(self, entry: TEntry):
        """Cancel a scheduled callback

        :param entry: the scheduler entry to be canceled
        """

    @abstractproperty
    def lock(self):
        """A global lock which can be used the assure thread safety"""


class AsyncScheduler(AbstractScheduler):
    def __init__(self, loop=None):
        self._loop = loop

        self._lock = threading.Lock()

    def add(self, delay: float, callback, args=(), kwargs=None) -> asyncio.TimerHandle:
        if kwargs:
            callback = functools.partial(callback, **kwargs)

        loop = self._loop or asyncio.get_event_loop()
        return loop.call_later(delay, callback, *args)

    def cancel(self, entry: asyncio.TimerHandle):
        entry.cancel()

    @property
    def lock(self):
        return self._lock


class SyncScheduler(AbstractScheduler):
    def __init__(self, lock: threading.Lock=None):
        if lock is None:
            lock = threading.Lock()
        self._lock = lock
        self._sched = scheduler()
        self._wake_up = threading.Event()

    def add(self, delay: float, callback, args=(), kwargs=None) -> asyncio.TimerHandle:
        if kwargs == None:
            kwargs = {}
        self._wake_up.set()
        return self._sched.enter(delay, 0, callback, args, kwargs)

    def cancel(self, entry):
        self._sched.cancel(entry)

    def run(self):
        while True:
            self._sched.run()
            self._wake_up.wait()
            self._wake_up.clear()

    @property
    def lock(self):
        return self._lock


class VirtualScheduler(AbstractScheduler):
    @dataclass
    class Entry:
        callback: Callable
        args: Tuple
        kwargs: Dict[str, Any]

    def __init__(self):
        self._time = 0
        self._lock = threading.Lock()
        self._id = 1
        self._entry_dict: Dict[int, VirtualScheduler.Entry] = dict()
        self._timestamp_dict: Dict[int, float] = dict()

    def add(self, delay: float, callback, args=(), kwargs=None) -> TEntry:
        if kwargs is None:
            kwargs = {}

        entry = VirtualScheduler.Entry(callback, args, kwargs)
        self._entry_dict[self._id] = entry
        self._timestamp_dict[self._id] = self._time + delay
        self._id += 1
        return self._id - 1

    def cancel(self, entry: int):
        self._entry_dict.pop(entry)
        self._timestamp_dict.pop(entry)

    def run(self, duration: float):
        start_time = self._time

        while self._entry_dict:
            earliest_timestamp = min(self._timestamp_dict.values())

            if earliest_timestamp > start_time + duration:
                break

            ids = tuple(
                id
                for id, timestamp in self._timestamp_dict.items()
                if timestamp == earliest_timestamp
            )
            for id in ids:
                entry = self._entry_dict[id]
                entry.callback(*entry.args, **entry.kwargs)
                self._entry_dict.pop(id)
                self._timestamp_dict.pop(id)

        self._time = start_time + duration

    @property
    def lock(self):
        return self._lock


_scheduler = AsyncScheduler()


def get_scheduler() -> AbstractScheduler:
    return _scheduler


def set_scheduler(scheduler: AbstractScheduler):
    global _scheduler
    _scheduler = scheduler



from durand.scheduler import get_scheduler, set_scheduler
from durand.node import MinimalNode, Node
from durand.object_dictionary import Variable, Record, Array
from durand.datatypes import DatatypeEnum


__all__ = [
    "MinimalNode",
    "Node",
    "Variable",
    "Record",
    "Array",
    "DatatypeEnum",
    "get_scheduler",
    "set_scheduler",
]



""" Testing TxPDOs """

from durand import Node, Variable
from durand.datatypes import DatatypeEnum as DT

from ..adapter import MockAdapter, TxMsg, RxMsg


def test_local_config_tpdo():
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    node.tpdo[0].mapping = [(0x2000, 0)]

    # send PDO message after changing into Operational state
    adapter.test(
        [   TxMsg(0x702, "00"),  # boot-up message from NMT

            RxMsg(0x000, "80 00"),  # set Pre-Operational state
            RxMsg(0x000, "01 00"),  # set Operational state

            TxMsg(0x182, "05 00")
        ]
    )

    # update the value and check, if PDO is sent
    node.object_dictionary.write(0x2000, 0, 0xAA)

    adapter.test(
        [   TxMsg(0x182, "AA 00")
        ]
    )

    # check PDO configuration via SDO
    adapter.test(
        [
            RxMsg(0x602, "40 00 18 01 00 00 00 00"),  # get cob id for sending
            TxMsg(0x582, "43 00 18 01 82 01 00 40"),  # receive 0x4000_0182
            RxMsg(0x602, "40 00 18 02 00 00 00 00"),  # get transmission type
            TxMsg(0x582, "4F 00 18 02 FF 00 00 00"),  # receive 0xFF
        ]
    )

    # disable PDO
    node.tpdo[0].enable = False

    node.object_dictionary.write(0x2000, 0, 0xBB)  # this will not trigger a PDO

    adapter.test(
        [
            RxMsg(0x602, "40 00 18 01 00 00 00 00"),  # get cob id for sending
            TxMsg(0x582, "43 00 18 01 82 01 00 C0"),  # receive 0xC000_0182
        ]
    )

def test_remote_config_tpdo():
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    assert node.tpdo[10].enable == False

    adapter.test(
        [   TxMsg(0x702, "00"),  # boot-up message from NMT

            RxMsg(0x602, "23 0A 1A 01 02 00 00 20"),  # map object 0x2000 to PDO 11
            TxMsg(0x582, "60 0A 1A 01 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x602, "2F 0A 1A 00 01 00 00 00"),  # set number of mapped objects to 1
            TxMsg(0x582, "60 0A 1A 00 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x602, "23 0A 18 01 00 02 00 40"),  # set cob id to 0x200 (and enable)
            TxMsg(0x582, "60 0A 18 01 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x000, "01 00"),  # set Operational state

            TxMsg(0x200, "05 00")
        ]
    )

    assert node.tpdo[10].enable == True
    assert node.tpdo[10].mapping == ((0x2000, 0),)


""" Testing RxPDOs """

from durand import Node, Variable
from durand.datatypes import DatatypeEnum as DT

from ..adapter import MockAdapter, TxMsg, RxMsg


def test_local_config_rpdo():
    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    node.rpdo[0].mapping = [(0x2000, 0)]

    # receive PDO message after changing into Operational state
    adapter.test(
        [   TxMsg(0x702, "00"),  # boot-up message from NMT

            RxMsg(0x000, "80 00"),  # set NMT Pre-Operational state
            RxMsg(0x202, "02 00")  # receive the PDO message (will be ignored)
        ]
    )

    assert node.object_dictionary.read(0x2000, 0) == 5

    adapter.test(
        [  
            RxMsg(0x000, "01 00"),  # set NMT Operational state

            RxMsg(0x202, "02 00")  # receive the PDO message
        ]
    )

    assert node.object_dictionary.read(0x2000, 0) == 2

    # check PDO configuration via SDO
    adapter.test(
        [
            RxMsg(0x602, "40 00 14 01 00 00 00 00"),  # get cob id for sending
            TxMsg(0x582, "43 00 14 01 02 02 00 00"),  # receive 0x0000_0202
            RxMsg(0x602, "40 00 14 02 00 00 00 00"),  # get transmission type
            TxMsg(0x582, "4F 00 14 02 FF 00 00 00"),  # receive 0xFF
        ]
    )

    # disable PDO
    node.rpdo[0].enable = False

    adapter.test([RxMsg(0x202, "01 00")])  # receive the PDO message
    
    assert node.object_dictionary.read(0x2000, 0) == 2
    
    adapter.test(
        [
            RxMsg(0x602, "40 00 14 01 00 00 00 00"),  # get cob id for sending
            TxMsg(0x582, "43 00 14 01 02 02 00 80"),  # receive 0x8000_0202
        ]
    )



from typing import TYPE_CHECKING
import struct

from durand.object_dictionary import Variable
from durand.services.nmt import StateEnum
from durand.datatypes import DatatypeEnum as DT
from durand.scheduler import get_scheduler

if TYPE_CHECKING:
    from durand.node import Node


class EMCY:
    def __init__(self, node: "Node"):
        self._node = node

        self._timer_handle = None
        self._inhibit_time = 0
        self._deferred_emcy = None

        self._cob_id = 0x80 + node.node_id

        node.object_dictionary[0x1001] = Variable(
            DT.UNSIGNED8, "ro", 0, name="Error Register"
        )
        node.object_dictionary[0x1014] = Variable(
            DT.UNSIGNED32, "rw", self._cob_id, name="COB-ID EMCY"
        )
        node.object_dictionary[0x1015] = Variable(
            DT.UNSIGNED16, "rw", self._inhibit_time, name="Inhibit Time EMCY"
        )

        node.object_dictionary.download_callbacks[(0x1014, 0)].add(
            self._downloaded_cob_id
        )
        node.object_dictionary.update_callbacks[(0x1015, 0)].add(
            self._update_inhibit_time
        )

        node.nmt.state_callbacks.add(self._update_nmt_state)

    def _downloaded_cob_id(self, value: int):
        self._cob_id = value

    def _update_nmt_state(self, state: StateEnum):
        if state == StateEnum.STOPPED:
            self._cob_id = None
        elif (
            state in (StateEnum.PRE_OPERATIONAL, StateEnum.OPERATIONAL)
            and self._cob_id is None
        ):
            self._cob_id = 0x80 + self._node.node_id

    @property
    def cob_id(self):
        return self._cob_id

    @cob_id.setter
    def cob_id(self, value: int):
        self._cob_id = value

    @property
    def enable(self) -> bool:
        return not self._cob_id & (1 << 31)

    @enable.setter
    def enable(self, value: bool):
        if value:
            self._cob_id &= ~(1 << 31)
        else:
            self._cob_id |= 1 << 31

        self._node.object_dictionary.write(0x1014, 0, self._cob_id, downloaded=False)

    def _update_inhibit_time(self, value: int):
        if self._timer_handle is not None:
            get_scheduler().cancel(self._timer_handle)
        self._timer_handle = None
        self._inhibit_time = value / 10_000

    @property
    def inhibit_time(self):
        return self._node.object_dictionary.read(0x1015, 0) / 10_000

    @inhibit_time.setter
    def inhibit_time(self, value: float):
        self._node.object_dictionary.write(0x1015, 0, value * 10_000, downloaded=False)

    def _time_up(self):
        self._timer_handle = None

        if self._deferred_emcy:
            self._send(*self._deferred_emcy)

        self._deferred_emcy = None

    def set(self, error_code: int, error_register: int, data: bytes = b""):
        self._node.object_dictionary.write(0x1001, 0, error_register, downloaded=False)

        if not self.enable:
            return

        if self._timer_handle is not None:
            self._deferred_emcy = (error_code, error_register, data)
            return

        self._send(error_code, error_register, data)

    def _send(self, error_code: int, error_register: int, data: bytes = b""):
        if self._inhibit_time:
            self._timer_handle = get_scheduler().add(self._inhibit_time, self._time_up)

        msg = (
            struct.pack("<HB", error_code, error_register) + data + bytes(5 - len(data))
        )
        self._node.adapter.send(self._cob_id, msg)



""" Testing TxPDOs with inhibt time """

from durand import Node, Variable, set_scheduler
from durand.scheduler import VirtualScheduler
from durand.datatypes import DatatypeEnum as DT

from ..adapter import MockAdapter, TxMsg, RxMsg


def test_local_config_inhibit():
    scheduler = VirtualScheduler()
    set_scheduler(scheduler)

    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    node.tpdo[0].mapping = [(0x2000, 0)]
    node.tpdo[0].inhibit_time = 0.5  # [s]

    # receive PDO message after changing into Operational state
    adapter.test(
        [   TxMsg(0x702, "00"),  # boot-up message from NMT

            RxMsg(0x000, "80 00"),  # set Pre-Operational state
            RxMsg(0x000, "01 00"),  # set Operational state

            TxMsg(0x182, "05 00")
        ]
    )

    # update the value and check, if PDO is sent (will be sent after 0.5s inhibit)
    node.object_dictionary.write(0x2000, 0, 0xAA)

    adapter.tx_mock.assert_not_called()

    scheduler.run(0.4)
    adapter.tx_mock.assert_not_called()

    scheduler.run(0.2)
    adapter.test(
        [   TxMsg(0x182, "AA 00")
        ]
    )

    scheduler.run(0.5)

    # update the value and check, if PDO is sent (will be sent immediatly)
    node.object_dictionary.write(0x2000, 0, 0xBB)

    adapter.test(
        [   TxMsg(0x182, "BB 00")
        ]
    )

    # read inhibt time via SDO
    adapter.test(
        [
            RxMsg(0x602, "40 00 18 03 00 00 00 00"),  # get inhibit time
            TxMsg(0x582, "4B 00 18 03 88 13 00 00"),  # receive 5000 [100µs]
        ]
    )


def test_remote_config_inhibit():
    scheduler = VirtualScheduler()
    set_scheduler(scheduler)

    adapter = MockAdapter()

    # create the node
    node = Node(adapter, node_id=2)

    # add a variable with index 0x2000 to the object dictionary of the node
    node.object_dictionary[0x2000] = Variable(DT.INTEGER16, "rw", value=5)

    node.tpdo[0].mapping = [(0x2000, 0)]

    adapter.test(
        [   TxMsg(0x702, "00"),  # boot-up message from NMT

            RxMsg(0x602, "2B 00 18 03 88 13 00 00"),  # set inhibit time to 5000 [100µs]
            TxMsg(0x582, "60 00 18 03 00 00 00 00"),  # response (acknowledge)

            RxMsg(0x000, "01 00"),  # set Operational state

            TxMsg(0x182, "05 00")
        ]
    )

    # update the value and change to NMT pre-operation state within inhibt time
    # PDO should not be sent
    node.object_dictionary.write(0x2000, 0, 0xAA)

    adapter.tx_mock.assert_not_called()

    scheduler.run(0.2)

    adapter.test(
        [   RxMsg(0x000, "80 00"),  # NMT go to pre-operational state
        ]
    )
    scheduler.run(1)

    adapter.tx_mock.assert_not_called()



from typing import TYPE_CHECKING

from durand.object_dictionary import Variable
from durand.datatypes import DatatypeEnum as DT
from durand.callback_handler import CallbackHandler


if TYPE_CHECKING:
    from durand.node import Node


class Sync:
    def __init__(self, node: "Node"):
        self._node = node
        self._cob_id = 0x80

        self.callbacks = CallbackHandler()

        node.object_dictionary[0x1005] = Variable(
            DT.UNSIGNED32, "rw", self._cob_id, name="COB-ID SYNC"
        )
        node.object_dictionary.update_callbacks[(0x1005, 0)].add(self._update_cob_id)

        node.adapter.add_subscription(cob_id=self._cob_id, callback=self._receive_sync)

    def _update_cob_id(self, value):
        self._node.adapter.remove_subscription(
            cob_id=self._cob_id, callback=self._receive_sync
        )
        self._cob_id = value & 0x1FFF_FFFF
        self._node.adapter.add_subscription(
            cob_id=self._cob_id, callback=self._receive_sync
        )

    def _receive_sync(self, cob_id: int, msg: bytes):
        self.callbacks.call()



import asyncio

from durand.object_dictionary import ObjectDictionary
from durand.datatypes import DatatypeEnum
from durand import MinimalNode, Variable
from durand.adapters.can import CANAdapter


adapter = CANAdapter(interface='socketcan', channel='can0', loop=asyncio.get_event_loop())

node = MinimalNode(adapter, 0x20)

state_control = Variable(DatatypeEnum.UNSIGNED8, access='rw', value=0)
node.object_dictionary[0x2000] = state_control

node.tpdo[1].mapping = ((0x2000, 0),)
node.rpdo[1].mapping = ((0x2000, 0),)

# change value
node.object_dictionary.write(0x2000, 0, 10)

# read value
node.object_dictionary.read(0x2000, 0)
asyncio.get_event_loop().call_later(2, node.object_dictionary.write, 0x2000, 0, 11)

data_var = Variable(DatatypeEnum.DOMAIN, access='rw')
node.object_dictionary[0x2001] = data_var

node.object_dictionary.update_callbacks[(0x2001, 0)].add(print)

class MyDownloadHandler:
    def __init__(self):
        self._file = open('update.bin', 'wb')

    def on_receive(self, data: bytes):
        self._file.write(data)

    def on_finish(self):
        self._file.close()

    def on_abort(self):
        self._file.close()

#define handler
def download_callback(node, index, subindex, size):
    if index == 0x2001:
        return MyDownloadHandler()

    return None

node.sdo_servers[0].download_manager.set_handler_callback(download_callback)

def change_baudrate(baudrate: int):
    print('CHANGE BAUDRATE', baudrate)

node.lss.set_baudrate_change_callback(change_baudrate)

asyncio.get_event_loop().run_forever()